// Author: Pete Kirkham
// tiny quad store benchmark.
//
// read in the SQL triple store dump, output as binary.

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <iostream>

#include "io/leap_frog_file_reader.h"
#include "utils/string_utils.h"
#include "sql/sql_reader.h"
#include "core/triple_target.h"

#include "lzw/lzw_compression.h"

class triple_compress : public triple_target
{
public:
    triple_compress (unsigned int heap_size, unsigned int long_strings_size = 360) : _compression(heap_size, true)
    {
        _compressed = new char[1024];
        _compressed_length = 0;
        _n_triples = 0;
        _very_compressed_strings = 0;
        _compressed_plus = 0;
        _hash_table = new unsigned int [16 * MiB];
        _long_strings = new char [long_strings_size * MiB];
        _long_strings_size = long_strings_size * MiB;
        _primary_collisions = 0;

        for (size_t i(0); i < 16 * MiB; ++i)
        {
            _hash_table[i] = -1;
        }

        _triples = new unsigned int[4 * 10 * MiB];
    }

    ~triple_compress ()
    {
        delete [] _compressed;
        delete [] _hash_table;
        delete [] _long_strings;
        delete [] _triples;
    }

    void add_triple (const char* subject,   size_t subject_length,   field_type subject_type,
                     const char* predicate, size_t predicate_length, // always an uri
                     const char* object,    size_t object_length,    field_type object_type,
                     const char* graph,     size_t graph_length)
    {
        unsigned int subject_id(process_string(subject, subject_length));
        unsigned int predicate_id(process_string(predicate, predicate_length));
        unsigned int object_id(process_string(object, object_length) | (object_type == type_literal ? 1 : 0));
        unsigned int graph_id(process_string(graph, graph_length));

        _triples[_n_triples << 2]     = subject_id;
        _triples[(_n_triples << 2)|1] = predicate_id;
        _triples[(_n_triples << 2)|2] = object_id;
        _triples[(_n_triples << 2)|3] = graph_id;

        ++_n_triples;
    }

    size_t compressed_length ()
    {
        return _compressed_length;
    }

    size_t n_triples ()
    {
        return _n_triples;
    }

    unsigned int* triples ()
    {
        return _triples;
    }

    size_t very_compressed_strings ()
    {
        return _very_compressed_strings;
    }

    size_t compressed_plus ()
    {
        return _compressed_plus;
    }

    size_t primary_collisions ()
    {
        return _primary_collisions;
    }

private:
    lzw::compression    _compression;
    char*               _compressed;
    size_t              _compressed_length;
    size_t              _n_triples;
    size_t              _very_compressed_strings;
    size_t              _compressed_plus;
    char*               _long_strings;
    size_t              _long_strings_size;
    unsigned int*       _hash_table;
    size_t              _primary_collisions;
    unsigned int*       _triples;

    unsigned int process_string (const char* input, size_t input_length)
    {
        char* compressed_string(_compressed);
        const size_t length(_compression.compress(input, input_length, compressed_string, 1024));

        _compressed_length += length;

        if ((length < 4) || ((length == 4) && (compressed_string[3]&0x03 == 0)))
        {
            ++_very_compressed_strings;

            return  ((compressed_string[0] & 0xff) << 24) |
                    ((compressed_string[1] & 0xff) << 16) |
                    ((compressed_string[2] & 0xff) << 8) |
                     (compressed_string[3] & 0xff);
        }
        else
        {
            // calculate 24 bit hash of string
            unsigned int hash(0);

            for (size_t i(0); i < length; ++i)
            {
                hash = hash * 63 ^ (0xff & compressed_string[i]);
            }

            unsigned int hash_24(mix(hash) & 0xffffff);

            // find corresponding index in hash table
            unsigned int index(_hash_table[hash_24]);
            unsigned int original_hash_24(hash_24);

            bool same(false);

            while (index != -1) // hash already in use, check whether string is repeated
            {
                // compare strings
                same = true;

                for (size_t i(0); i <= length; ++i)
                {
                    if (_long_strings[index + i] != compressed_string[i])
                    {
                        same = false;
                        break;
                    }
                }

                if (same) // string is repeat of hashed string.
                {
                    break;
                }

                if (hash_24 == original_hash_24)
                {
                    ++_primary_collisions;
                }

                // increment hash_24 and try again.
                hash_24 = (hash_24 + 1) & 0xffffff;

                // unable to resolve collision
                if (hash_24 == original_hash_24)
                {
                    std::cerr << "unable to resolve hash collision in string table.";
                    exit(-1);
                }

                index = _hash_table[hash_24];
            }

            if (index == -1) // unused hash, copy string to long string store
            {
                index = _compressed_plus;

                _compressed_plus += length + 1;

                if (_compressed_plus < _long_strings_size)
                {
                    _hash_table[hash_24] = index;

                    for (size_t i(0); i <= length; ++i)
                    {
                        _long_strings[index + i] = compressed_string[i];
                    }
                }
            }
            else if (!same)
            {
                std::cout << "collision index: " << index << " hash: " << hash_24 << std::endl;
            }

            return (index << 1) | 0x01;
        }
    }
};

void check_collision (size_t term_0, size_t term_1, unsigned int* triples, size_t i, size_t* hash_table, size_t mask, size_t* occurance_table, size_t& close_collisions, size_t& far_collisions)
{
    unsigned int v0(triples[(i<<2)|term_0]);
    unsigned int v1(triples[(i<<2)|term_1]);
    unsigned int hash_index((mask & mix(v0/* ^ mix(v1)*/)) << 1);

    if (hash_table[hash_index] == 0)
    {
        hash_table[hash_index] = v0;
        hash_table[hash_index|1] = v1;
    }
    else if (hash_table[hash_index] != v0)
    {
        if (far_collisions < 16)
        {
            std::cout << "Far collision @ " << to_hex(hash_index>>1) << " between q_" << term_0 << ',' << term_1 << " = </" << v0 << ',' << v1 << "/> and </" << hash_table[hash_index] << ',' << hash_table[hash_index|1] << "/>" << std::endl;
        }

        ++far_collisions;
    }
    else if (hash_table[hash_index|1] != v1)
    {
        if (close_collisions < 16)
        {
            std::cout << "Close collision @ " << to_hex(hash_index>>1) << " between q_" << term_0 << ',' << term_1 << " = </" << v0 << ',' << v1 << "/> and </" << hash_table[hash_index] << ',' << hash_table[hash_index|1] << "/>" << std::endl;
        }

        ++close_collisions;
    }

    ++occurance_table[hash_index>>1];
}

int main (int narg, char** argvp)
{
    if (narg <  2)
    {
        std::cerr << "The input file name is required. Optionally, the number of MiB to read and the compression dictionary size in KiB." << std::endl;
        return 1;
    }

    size_t max_read(narg > 2 ? atoi(argvp[2]) * MiB : std::numeric_limits<size_t>::max());
    size_t heap_size(narg > 3 ? atoi(argvp[3]) * KiB : 32 * MiB);

    // open the file
    int file(open(argvp[1], O_RDONLY));

    if (!file)
    {
        std::cerr << "failed to open " << argvp[1] << std::endl;
        return 2;
    }

    leap_frog_file_reader                 fr(file);

    // skip the first 1270 bytes so we're in the SQL dump, and can super::expect to find "INSERT INTO `triple_all` VALUES"
    if (!fr.seek(1270))
    {
        std::cerr << "error searching for table start" << std::endl;
        exit(1);
    }

    sql_reader<leap_frog_file_reader>     reader(fr);
    triple_compress             target(heap_size);

    // read in data
    reader.write_to(target, max_read);

    if (reader.total_read() < 8 * MiB)
    {
        std::cout << "#read       " << reader.total_read()/KiB << " KiB." << std::endl;
    }
    else
    {
        std::cout << "#read       " << reader.total_read()/MiB << " MiB." << std::endl;
    }

    if (target.compressed_length() < 8 * MiB)
    {
        std::cout << "#compressed " << target.compressed_length()/KiB << " KiB." << std::endl;
    }
    else
    {
        std::cout << "#compressed " << target.compressed_length()/MiB << " MiB." << std::endl;
    }

    std::cout << "# " << target.very_compressed_strings() << " strings compressed into 30 bits each" << std::endl;

    if (target.compressed_plus() < 8 * MiB)
    {
        std::cout << "#the remainder, compressed into " << target.compressed_plus()/KiB << " KiB." << std::endl;
    }
    else
    {
        std::cout << "#the remainder, compressed into " << target.compressed_plus()/MiB << " MiB." << std::endl;
    }

    close(file);
    
#ifdef FIND_HASHES
    // find collisions in hashes
    size_t          hash_table_size(1<<22);
    size_t          mask(hash_table_size-1);
    size_t*         hash_table(new unsigned int[hash_table_size*2]);
    size_t*         occurance_table(new unsigned int[hash_table_size]);
    unsigned int*   triples(target.triples());
    size_t          n_triples(target.n_triples());

    for (size_t term_0(0); term_0 < 3; ++term_0)
    {
        for (size_t term_1(term_0+1); term_1 < 4; ++term_1)
        {
            size_t          close_collisions(0);
            size_t          far_collisions(0);

            for (size_t i(0); i < hash_table_size*2; ++i)
            {
                hash_table[i] = 0;
            }

            for (size_t i(0); i < hash_table_size; ++i)
            {
                occurance_table[i] = 0;
            }

            for (size_t i(0); i < n_triples; ++i)
            {
                check_collision(term_0, term_1, triples, i, hash_table, mask, occurance_table, close_collisions, far_collisions);
            }

            size_t max_occupancy(0);
            size_t occupied_buckets(0);

            for (size_t i(0); i < hash_table_size; ++i)
            {
                if (occurance_table[i] > max_occupancy) max_occupancy = occurance_table[i];
                if (occurance_table[i]) ++occupied_buckets;
            }

            std::cout << "hashing q_" << term_0 << ',' << term_1 << " into " << hash_table_size <<
                " generated " << close_collisions << " close collisions and " << far_collisions << " far collisions." << std::endl;
            std::cout << " of the " << occupied_buckets << " non-empty hash buckets, the largest occupancy is " << max_occupancy << "." << std::endl;
        }
    }


    delete [] hash_table;
#endif // FIND_HASHES
    return 0;
}
 
